/*
 * gMusic Score Editor
 *
 * Copyright (C) 1995-2001 Allan R. Black
 *
 * gMusic/choice.c
 *
 */

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <gtkchoice.h>

#include <gdk/gdkprivate.h>

enum {
	SELECTION,
	LAST_SIGNAL
};

static void	gtk_choice_class_init(GtkChoiceClass *);
static void	gtk_choice_init(GtkChoice *);
/*static void	gtk_choice_realize(GtkWidget *);*/
static void	gtk_choice_destroy(GtkObject *);
static void	gtk_real_choice_select(GtkChoice *, guint);

static guint	choice_signals[LAST_SIGNAL];

GtkType gtk_choice_get_type(void)
{
	static guint choice_type = 0;

	if(!choice_type) {
		static const GtkTypeInfo choice_info = {
			"GtkChoice",
			sizeof(GtkChoice),
			sizeof(GtkChoiceClass),
			(GtkClassInitFunc) gtk_choice_class_init,
			(GtkObjectInitFunc) gtk_choice_init,
			NULL, NULL,
			(GtkClassInitFunc)0
		};

		choice_type = gtk_type_unique(gtk_button_get_type(),
						&choice_info);
	}

	return choice_type;
}

static void gtk_choice_class_init(GtkChoiceClass *klass)
{
	GtkObjectClass *oclass;
	/*GtkWidgetClass *wclass;*/

	oclass = (GtkObjectClass *)klass;
	/*wclass = (GtkWidgetClass *)klass;*/

	choice_signals[SELECTION] = gtk_signal_new("select",
			GTK_RUN_FIRST, oclass->type,
			GTK_SIGNAL_OFFSET(GtkChoiceClass, choice_select),
			gtk_marshal_NONE__UINT, GTK_TYPE_NONE,
			1, GTK_TYPE_UINT);
	gtk_object_class_add_signals(oclass, choice_signals, LAST_SIGNAL);

	/*wclass->realize = gtk_choice_realize;*/
	klass->choice_select = gtk_real_choice_select;
}

static void gtk_real_choice_select(GtkChoice *choice, guint selection)
{
	GtkListItem *lchild;
	gchar *text;
	guint16 width;

	g_return_if_fail(choice != NULL);
	g_return_if_fail(selection < choice->nitems);

	lchild = g_list_nth_data(GTK_LIST(choice->list)->children, selection);
	g_return_if_fail(lchild != NULL);

	gtk_label_get(GTK_LABEL(GTK_BIN(lchild)->child), &text);
	if(text == NULL) text = "";
	width = GTK_WIDGET(choice)->allocation.width;
	gtk_label_set_text(GTK_LABEL(choice->label), text);
	if(width > 1) gtk_widget_set_usize(GTK_WIDGET(choice), width, 0);

	choice->selection = selection;
	choice->has_selection = 1;
}

static void gtk_choice_select(GtkChoice *choice, guint selection)
{
	gtk_signal_emit(GTK_OBJECT(choice), choice_signals[SELECTION],
							selection);
}

static void gtk_choice_popdown(GtkChoice *choice)
{
	GdkEventCrossing synth;

	gtk_grab_remove(choice->list);
	gtk_grab_remove(choice->menu);
	gtk_widget_hide(choice->popup);

	choice->active = 0;
	choice->in_menu = 0;
	choice->in_list = 0;
	gtk_button_released(GTK_BUTTON(choice));
	gtk_grab_remove(GTK_WIDGET(choice));

	if(!choice->in_button) {
		synth.type = GDK_LEAVE_NOTIFY;
		synth.window = GTK_WIDGET(choice)->window;
		synth.send_event = TRUE;
		synth.subwindow = NULL;
		synth.detail = GDK_NOTIFY_ANCESTOR;
		gtk_widget_event(GTK_WIDGET(choice), (GdkEvent *)&synth);
	}
}

static gint gtk_choice_button_press
(
	GtkWidget	*w,
	GdkEventButton	*bu,
	GtkChoice	*choice
)
{
	GtkRequisition menureq;
	GtkAllocation popupall;
	GList *listsel;
	gint x, y;
	guchar hpolicy, vpolicy;

printf("CALLED: gtk_choice_button_press, w = 0x%x\n", w);
	if(bu->button != 1) return FALSE;

	gdk_window_get_origin(GTK_WIDGET(choice)->window, &x, &y);
printf("BUTTON POSITION: (%d, %d)\n", x, y);
printf("ALLOCATION: (%d, %d)\n", GTK_WIDGET(choice)->allocation.width, GTK_WIDGET(choice)->allocation.height);
	y += GTK_WIDGET(choice)->allocation.height;
printf("POPUP POSITION: (%d, %d)\n", x, y);
	gtk_widget_set_uposition(choice->popup, x, y);

	hpolicy = GTK_POLICY_NEVER;
	vpolicy = GTK_POLICY_NEVER;

printf("LIST REQUISITION: %d, %d\n", choice->list->requisition.width, choice->list->requisition.height);
printf("LIST ALLOCATION: %d, %d\n", choice->list->allocation.width, choice->list->allocation.height);
	gtk_widget_size_request(choice->menu, &menureq);
	/*gtk_widget_size_allocate(choice->menu, &popupall);*/
printf("LIST REQUISITION: %d, %d\n", choice->list->requisition.width, choice->list->requisition.height);
printf("LIST ALLOCATION: %d, %d\n", choice->list->allocation.width, choice->list->allocation.height);
printf("MENUREQ: %d, %d\n", menureq.width, menureq.height);
printf("POPUPALL: %d, %d\n", popupall.width, popupall.height);
printf("MENU REQUISITION: %d, %d\n", choice->menu->requisition.width, choice->menu->requisition.height);
printf("MENU ALLOCATION: %d, %d\n", choice->menu->allocation.width, choice->menu->allocation.height);
	if(choice->list->requisition.width > menureq.width) {
		gtk_widget_size_request(choice->menu, &menureq);
printf("---------- RESIZING ----------\n");
printf("LIST REQUISITION: %d, %d\n", choice->list->requisition.width, choice->list->requisition.height);
printf("LIST ALLOCATION: %d, %d\n", choice->list->allocation.width, choice->list->allocation.height);
printf("MENUREQ: %d, %d\n", menureq.width, menureq.height);
printf("MENU REQUISITION: %d, %d\n", choice->menu->requisition.width, choice->menu->requisition.height);
printf("MENU ALLOCATION: %d, %d\n", choice->menu->allocation.width, choice->menu->allocation.height);
	}

	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(choice->menu),
				GTK_POLICY_AUTOMATIC, vpolicy);
	gtk_widget_set_usize(choice->popup,
			GTK_WIDGET(choice)->allocation.width, 0);
	gtk_widget_show(choice->popup);
	listsel = GTK_LIST(choice->list)->selection;
	if(listsel)
		gtk_list_unselect_child(GTK_LIST(choice->list), listsel->data);
	gtk_grab_add(choice->menu);

	choice->active = 1;
	choice->dragging = 1;
	choice->in_button = 1;
	choice->has_selection = 0;

	return TRUE;
}

static gint gtk_choice_button_release
(
	GtkWidget	*w,
	GdkEventButton	*bu,
	GtkChoice	*choice
)
{
	if(choice->in_button && choice->dragging) {
		choice->dragging = 0;
		gdk_pointer_grab(choice->menu->window, TRUE,
				GDK_POINTER_MOTION_HINT_MASK
				| GDK_BUTTON1_MOTION_MASK
				| GDK_BUTTON_RELEASE_MASK,
				NULL, NULL, bu->time);
		return;
	}

	if(choice->in_list && choice->has_selection)
		gtk_choice_select(choice, choice->selection);
	if(choice->in_list || !choice->in_menu) gtk_choice_popdown(choice);
}

static void gtk_choice_selection
(
	GtkWidget	*w,
	GtkChoice	*choice
)
{
	GList *listsel;
	guint lpos;

	listsel = GTK_LIST(choice->list)->selection;
	if(listsel) {
		lpos = gtk_list_child_position(GTK_LIST(choice->list),
						listsel->data);
		if(lpos < choice->nitems) {
			choice->has_selection = 1;
			choice->selection = lpos;
		}
	}
}

static void gtk_choice_event_enter
(
	GtkWidget		*w,
	GdkEventCrossing	*cr,
	GtkChoice		*choice
)
{
	GtkWidget *ew;
	GList *listsel;

	/*
	 * We're not interested in moving out of a child window, nor in enter
	 * events generated as a result of grabs.
	 */
	if(cr->detail == GDK_NOTIFY_INFERIOR
		|| cr->mode != GDK_CROSSING_NORMAL) return;

	ew = gtk_get_event_widget((GdkEvent *)cr);
	if(ew == (GtkWidget *)choice)
		choice->in_button = 1;
	else if(ew == choice->popup)
		choice->in_menu = 1;
	else if(ew == choice->list) {
		gtk_grab_add(choice->list);
		choice->in_list = 1;
		GTK_LIST(choice->list)->drag_selection = TRUE;
		gdk_pointer_grab(choice->list->window, TRUE,
				GDK_POINTER_MOTION_HINT_MASK
				| GDK_BUTTON1_MOTION_MASK
				| GDK_BUTTON_RELEASE_MASK,
				NULL, NULL, cr->time);
	}
}

static void gtk_choice_event_leave
(
	GtkWidget		*w,
	GdkEventCrossing	*cr,
	GtkChoice		*choice
)
{
	GtkWidget *ew;
	GList *listsel;

	if(w == (GtkWidget *)choice && choice->active) {
		/*
		 * If a leave event is delivered to the button, then it's
		 * because gtk_main_do_event() has set the private flag
		 * GTK_LEAVE_PENDING in the widget. This, unfortunately,
		 * overrides the grab on the popup window and delivers the
		 * leave event to the button. So, in order to prevent the
		 * button reacting to pointer movement while the menu is
		 * active, we must explicitly stop the event. (Strangely
		 * enough, this problem doesn't manifest itself if the
		 * initial button press is made in the event box.)
		 */
		choice->in_button = 0;
		gtk_signal_emit_stop_by_name(GTK_OBJECT(choice),
						"leave_notify_event");
		return;
	}

	/*
	 * We're not interested in moving into a child window (i.e. a
	 * GtkListItem or the event box in the button), nor are we interested
	 * in leave events generated as a result of ungrabs.
	 */
	if(cr->detail == GDK_NOTIFY_INFERIOR
		|| cr->mode != GDK_CROSSING_NORMAL) return;

	ew = gtk_get_event_widget((GdkEvent *)cr);
	if(ew == (GtkWidget *)choice)
		choice->in_button = 0;
	else if(ew == choice->popup)
		choice->in_menu = 0;
	else if(ew == choice->list) {
		choice->in_list = 0;
		gtk_list_end_drag_selection(GTK_LIST(choice->list));
		gdk_pointer_grab(choice->menu->window, TRUE,
				GDK_POINTER_MOTION_HINT_MASK
				| GDK_BUTTON1_MOTION_MASK
				| GDK_BUTTON_RELEASE_MASK,
				NULL, NULL, cr->time);
		listsel = GTK_LIST(choice->list)->selection;
		if(listsel)
			gtk_list_unselect_child(GTK_LIST(choice->list),
							listsel->data);
	}
}

static void gtk_choice_button_press_cancel
(
	GtkWidget	*w,
	GdkEventButton	*bu,
	GtkChoice	*choice
)
{
if(choice->active && !choice->dragging && !choice->in_list) {
printf("CANCEL!\n");
printf("LIST REQUISITION: %d, %d\n", choice->list->requisition.width, choice->list->requisition.height);
printf("LIST ALLOCATION: %d, %d\n", choice->list->allocation.width, choice->list->allocation.height);
printf("MENU REQUISITION: %d, %d\n", choice->menu->requisition.width, choice->menu->requisition.height);
printf("MENU ALLOCATION: %d, %d\n", choice->menu->allocation.width, choice->menu->allocation.height);
}
	if(choice->active && !choice->dragging && !choice->in_menu)
		gtk_choice_popdown(choice);
}

static void gtk_choice_init(GtkChoice *choice)
{
	GtkWidget *hbox, *frame, *w;
	GtkStyle *style;
	GdkCursor *cursor;
	int i;

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(choice), hbox);
	gtk_container_set_border_width(GTK_CONTAINER(choice), 0);

	choice->label = gtk_label_new("");
	gtk_widget_set_usize(choice->label, 150, 0); /* ?? */
	gtk_misc_set_alignment(GTK_MISC(choice->label), 0.0f, 0.0f);
	GTK_WIDGET_UNSET_FLAGS(choice->label, GTK_CAN_FOCUS);
	gtk_misc_set_padding(GTK_MISC(choice->label), 2, 2);
	frame = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
	w = gtk_event_box_new();
	gtk_widget_ensure_style(w);
	style = gtk_style_copy(gtk_widget_get_style(w));
	for(i = 0; i < 5; ++i) style->bg[i] = style->base[i];
	gtk_widget_set_style(w, style);
	gtk_container_add(GTK_CONTAINER(w), choice->label);
	gtk_container_add(GTK_CONTAINER(frame), w);
	gtk_box_pack_start(GTK_BOX(hbox), frame, TRUE, TRUE, 0);

	choice->arrow = gtk_arrow_new(GTK_ARROW_DOWN, GTK_SHADOW_OUT);
	gtk_box_pack_start(GTK_BOX(hbox), choice->arrow, FALSE, FALSE, 0);

	gtk_signal_connect(GTK_OBJECT(choice), "leave_notify_event",
				GTK_SIGNAL_FUNC(gtk_choice_event_leave),
				choice);
	gtk_signal_connect_after(GTK_OBJECT(choice), "button_press_event",
				GTK_SIGNAL_FUNC(gtk_choice_button_press),
				choice);

	choice->popup = gtk_window_new(GTK_WINDOW_POPUP);
	gtk_widget_add_events(choice->popup, GDK_BUTTON_PRESS_MASK
						| GDK_BUTTON_RELEASE_MASK);
	frame = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_OUT);
	choice->menu = gtk_scrolled_window_new(NULL, NULL);
	gtk_signal_connect(GTK_OBJECT(choice->menu), "button_press_event",
				GTK_SIGNAL_FUNC(gtk_choice_button_press_cancel),
				choice);
	gtk_signal_connect(GTK_OBJECT(choice->menu), "button_release_event",
				GTK_SIGNAL_FUNC(gtk_choice_button_release),
				choice);
	gtk_signal_connect(GTK_OBJECT(choice->menu), "enter_notify_event",
				GTK_SIGNAL_FUNC(gtk_choice_event_enter),
				choice);
	gtk_signal_connect(GTK_OBJECT(choice->menu), "leave_notify_event",
				GTK_SIGNAL_FUNC(gtk_choice_event_leave),
				choice);
	choice->list = gtk_list_new();
	gtk_widget_add_events(choice->list, GDK_ENTER_NOTIFY_MASK
						| GDK_LEAVE_NOTIFY_MASK);
	gtk_list_set_selection_mode(GTK_LIST(choice->list),
				GTK_SELECTION_BROWSE);
	gtk_signal_connect(GTK_OBJECT(choice->list), "enter_notify_event",
				GTK_SIGNAL_FUNC(gtk_choice_event_enter),
				choice);
	gtk_signal_connect(GTK_OBJECT(choice->list), "leave_notify_event",
				GTK_SIGNAL_FUNC(gtk_choice_event_leave),
				choice);
	gtk_signal_connect(GTK_OBJECT(choice->list), "selection_changed",
				GTK_SIGNAL_FUNC(gtk_choice_selection), choice);
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(choice->menu),
						choice->list);
	gtk_container_add(GTK_CONTAINER(frame), choice->menu);
	gtk_container_add(GTK_CONTAINER(choice->popup), frame);

	gtk_widget_realize(choice->popup);
	cursor = gdk_cursor_new(GDK_ARROW);
	gdk_window_set_cursor(choice->popup->window, cursor);
	gdk_cursor_destroy(cursor);
	gtk_widget_show_all(frame);

	choice->nitems = 0;
	choice->has_selection = 0;
	choice->active = 0;
	choice->dragging = 0;
	choice->in_button = 0;
	choice->in_menu = 0;
	choice->in_list = 0;
}

/*static void gtk_choice_realize(GtkWidget *w)
{
printf("gtk_choice_realize\n");
	g_return_if_fail(w != NULL);
	g_return_if_fail(GTK_IS_CHOICE(w));
}*/

GtkWidget *gtk_choice_new(void)
{
	return GTK_WIDGET(gtk_type_new(gtk_choice_get_type()));
}

void gtk_choice_menu_clear(GtkChoice *choice)
{
	gtk_list_clear_items(GTK_LIST(choice->list), 0, 0);
}

void gtk_choice_menu_add(GtkChoice *choice, gchar *item)
{
	GtkWidget *w;

	w = gtk_list_item_new_with_label(item);
	gtk_container_add(GTK_CONTAINER(choice->list), w);
	gtk_widget_show(w);
	choice->nitems = gtk_list_child_position(GTK_LIST(choice->list), w)+1;
}

void gtk_choice_menu_set(GtkChoice *choice, GList *items)
{
	GList *l;

	gtk_choice_menu_clear(choice);
	for(l = items; l; l = l->next) gtk_choice_menu_add(choice, l->data);
}

void gtk_choice_menu_select(GtkChoice *choice, guint selection)
{
	if(selection < choice->nitems)
		gtk_choice_select(choice, selection);
}

void gtk_choice_menu_unselect(GtkChoice *choice)
{
	gtk_label_set_text(GTK_LABEL(choice->label), "");
	choice->has_selection = 0;
}

